"""
This file contains all the widgets required for use by the Fidelity Chartbook2001 update.
Each widget is called by the appropriate drawing class in the separate .py file with the
same name, which also inherits from DataAwareDrawing and uses that to make these widgets
both 'active' and 'data-aware'.

eg.
the class SlideBox in this file is used by the class SlideBoxDrawing in the file slidebox.py
"""
from types import FunctionType
from pprint import pprint
from reportlab.graphics.charts.textlabels import Label
from reportlab.graphics.shapes import Polygon, Line, Circle, String, Drawing, PolyLine, Group, Rect
from reportlab.graphics.charts.legends import Legend
from reportlab.graphics.charts.lineplots import LinePlot
from reportlab.graphics.charts import piecharts
from reportlab.graphics.widgetbase import Widget, TypedPropertyCollection
from reportlab.graphics.widgets.grids import ShadedRect, Grid
from reportlab.lib import colors
from reportlab.lib.attrmap import *
from reportlab.lib.formatters import Formatter
from reportlab.lib.units import cm
from reportlab.lib.validators import *
from reportlab.lib.utils import flatten
from reportlab.pdfbase.pdfmetrics import stringWidth, getFont
from reportlab.graphics.charts import barcharts
import fidlineplot
from rlextra.graphics.guiedit.datacharts import DataAwareDrawing, CSVDataSource
from reportlab.graphics.shapes import _DrawingEditorMixin

from fidcolors import *
from fidcolors import _PCMYK_black
###
## Helper classes and functions
###




def _maxWidth(text, fontName, fontSize):
    #applies stringWidth to the list text
    ##return max(map(lambda t, fontName=fontName, fontSize=fontSize: stringWidth(t, fontName, fontSize), text))
    t=[]
    for f in range(0,len(text)):
        if text[f]!=None and max(text)!=None:
            t.append(stringWidth(text[f], fontName, fontSize))
    if len(t)>0:
        return max(t)
    else:
        return 0

def _highest(list, roundInterval=10):
    v=max(list(map(max, list)))
    d,f=divmod(v,roundInterval)
    if f>0:
        f=1
    else:
        f=0.5
    return (d+f)*roundInterval


class DRRDrawing(DataAwareDrawing):
    """This contains behaviour shared by all the DRR charts
    
    They all read the same style of CSV files and should be able to strip out
    unwanted funds.
    
    We also provide 'helper functions' which can be used in the dynamically
    calculated 'getContents' section of each chart, such as standard number
    formatting, oir calculating output directories.
    
    """
    _attrMap = AttrMap(BASE=DataAwareDrawing,
        language = AttrMapValue(isString, desc="Language code"),
        )
    
    def __init__(self,width=400,height=200,*args,**kw):
        DataAwareDrawing.__init__(self,width,height,*args,**kw)
        self.language = 'en'
        self.dataSource = CSVDataSource()
        self.dataSource.headerRows = 1
        self.dataSource.sep = '\t'
        self.dataSource.csv_reader_kwds = dict(delimiter='\t')
        #self.dataSource.ignoreLinesMatching = '^FM'

    def suffixFilenameWithLanguage(self):
        """want to end with language"""
        suffix = '_' + self.language
        if not self.fileNamePattern.endswith(suffix):
            self.fileNamePattern += suffix
        
    def applyDataSet(self, dataSet):
        """Enforce the filename for all charts"""
        self.suffixFilenameWithLanguage()
        return DataAwareDrawing.applyDataSet(self, dataSet)
        
    def fmtDecimal1(self,num):
        """Language-specific formatting:   8.1 in english, 8,1 in French
        
        We plan on using these language codes:
            en = English
            xx = generic international, use ',' for decimal point
            fr = French
            nl = Dutch
        """
        txt = '%0.1f' % num
        if self.language != 'en':
            txt = txt.replace('.',',')
        return txt
    
    def fmtDecimal2(self,num):
        """Language-specific formatting:   8.1 in english, 8,1 in French
        
        We plan on using these language codes:
            en = English
            xx = generic international, use ',' for decimal point
            fr = French
            nl = Dutch
        """
        txt = '%0.2f' % num
        if self.language != 'en':
            txt = txt.replace('.',',')
        return txt

    def fmtPercent1(self, num):
        """Language-specific formatting:   8.1% in english, 8,1% in French
        
        We plan on using these language codes:
            en = English
            xx = generic international, use ',' for decimal point
            fr = French
            nl = Dutch
         
        (We can't use 'int', which is the name of a Python function and it
        confuses the system passing things in on the command line).
        Any formatting functions we write should take account of this, and
        may need reviewing if more languages are added.
        """
        txt = '%0.1f%%' % num
        if self.language != 'en':
            txt = txt.replace('.',',')
        return txt

    def fmtPercent2(self, num):
        """Format percent with two decimal places e.g. 3.15%"""
        txt = '%0.2f%%' % num
        if self.language != 'en':
            txt = txt.replace('.',',')
        return txt

    def fmtPercent3(self, num):
        """Format percent with three decimal places e.g. 3.152%"""
        txt = '%0.2f%%' % num
        if self.language != 'en':
            txt = txt.replace('.',',')
        return txt

    def fmtPercent0(self, num):
        "Format percent with no decimal  3.152 -> 3%"
        txt = '%df%%' % num
        if self.language != 'en':
            txt = txt.replace('.',',')
        return txt
    
        
    
class LineChart(fidlineplot.FidLinePlot):
    """Returns a widget - used by the class LineChartDrawing.
    Used for Chartbook 2001 - small diagram which appears under eg 'Performace Since Launch'.
    This class is basically a version of FidLineDrawing with the size, colours and other
    defaults set to be those used in the 2001 Chartbook.
    """
    _attrMap = AttrMap(BASE=fidlineplot.FidLinePlot,
        legend = AttrMapValue(None, desc="Handle for chart legend"),
        leftPadding = AttrMapValue(isNumber,
            desc='Padding on left of drawing'),
        rightPadding = AttrMapValue(isNumber,
            desc='Padding on right of drawing'),
        topPadding = AttrMapValue(isNumber,
            desc='Padding at top of drawing'),
        bottomPadding = AttrMapValue(isNumber,
            desc='Padding at bottom of drawing'),
        )
    
    def __init__(self):
        fidlineplot.FidLinePlot.__init__(self)
        self.x = 20
        self.y = 30
        self.height = 3.1*cm
        self.width = 6.1*cm
        #self.strokeColor = _PCMYK_black
        
        c0 = ChbkBlue # Fund
        self.lines[0].strokeColor = c0
        self.lines[0].strokeWidth = 2
        self.lines[0].strokeDashArray = None
        
        c1 = ChbkYellow # Index
        self.lines[1].strokeColor = c1
        self.lines[1].strokeWidth = 2
        self.lines[1].strokeDashArray = None
        from sampledata import monthlyIndexData
        self.data=monthlyIndexData
        
        self.xValueAxis.visibleAxis = 0
        self.xValueAxis.visibleTicks = 0
        self.xValueAxis.labels.fontSize = 6
        self.xValueAxis.labels.fontName = "Avenir-Light"
        self.xValueAxis.labels.textAnchor = 'start'
        self.xValueAxis.labels.boxAnchor = 'e'
        self.xValueAxis.labels.angle = 45
        self.xValueAxis.xLabelFormat = '{mmm} {yy}'
        
        self.yValueAxis.valueMin = 100.0
        self.yValueAxis.valueMax = None
        
        self.yValueAxis.visibleAxis = 0
        self.yValueAxis.visibleTicks = 0
        self.yValueAxis.labels.fontSize = 6
        self.yValueAxis.labels.fontName = "Avenir-Light"
        self.yValueAxis.labelTextFormat = None
        
        self.legend=Legend()
        self.legend.fontName = "Avenir-Light"
        self.legend.fontSize = 6.5
        self.legend.deltax = 1.2*cm # legendXDistance
        self.legend.deltay = 5 # legendYDistance
        self.legend.dxTextSpace = 0.12*cm # legendTextSpace
        self.legend.dx = 5 # legendSwatchWidth
        self.legend.dy = 5 # legendSwatchHeight
        self.legend.alignment = 'right' # legendSwatchAlign
        self.legend.colorNamePairs = [(c0, "Fund"), (c1, "Index")]
        self.legend.strokeColor = None # legendBorderColor = None
        
        #self.background = Grid()
        #self.background.useLines = 1
        #self.background.useRects = 0
        #self.background.strokeColor = _PCMYK_black
        #self.background.strokeWidth = 0.5
        #self.background.orientation = 'horizontal'
        #self.background.rectStrokeColor = None
        
        self.leftPadding = 5
        self.rightPadding = 5
        self.topPadding = 5
        self.bottomPadding = 5
        
        self.legend.columnMaximum = 1
        self.legend.x = (self.width+self.x)-1.25*(self.legend.deltax+self.legend.dxTextSpace+self.legend.dx)
        self.legend.x = self.legend.x-_maxWidth(self.legend.colorNamePairs[-1][1], self.legend.fontName, self.legend.fontSize)
        self.legend.y = self.bottomPadding+(self.legend.fontSize/100.0)*25
    
    def draw(self):
        #basic LinePlot - does the Axes, Ticks etc
        p=fidlineplot.FidLinePlot.draw(self)
        p.add(self.legend)
        p.shift(self.leftPadding, self.bottomPadding)
        return p
    
    def demo(self, drawing=None):
        if not drawing:
            drawing = Drawing(205, 130)
            #drawing = Drawing(400, 200)
        drawing.add(self.draw())
        return drawing

def _addLegendData(legend,f,data,lNF):
    if data != None:
        ldf = data[f]
        if ldf is None or lNF is None:
            pass
        elif type(lNF) is str:
            ldf = lNF % ldf
        elif type(lNF) is FunctionType:
            ldf = lNF(ldf)
        elif isinstance(lNF, Formatter):
            ldf = lNF(ldf)
        else:
            raise ValueError("Unknown formatter type %s, expected string or function" % type(lNF))
        legend.colorNamePairs.append((None,ldf))

class PieChart(piecharts.Pie):
    """Returns a widget - used by the class PieChartDrawing (in the file piechart.py).
    Used for Chartbook 2001 - small diagram which appears under eg 'Credit Rating'.
    This class uses piecharts.Pie as a basis, then adds a two part legend (one
    editable with swatches, one hidden without swatches).
    """
    
    _attrMap = AttrMap(BASE=piecharts.Pie,
        drawLegend = AttrMapValue(isBoolean, desc="If true then create and draw legend"),
        legend1 = AttrMapValue(None, desc="Handle to legend for pie"),
        legendNumberFormat = AttrMapValue(None, desc="Formatting routine for percentage on right hand side of legend."),
        legendNumberOffset = AttrMapValue(isNumber, desc="Horizontal space between legend and percentage on r/hand side"),
        pieAndLegend_colors = AttrMapValue(isListOfColors, desc="Colours used for both swatches and pie"),
        legend_names = AttrMapValue(isNoneOrListOfNoneOrStrings, desc="Names used in legend (or None)"),
        legend_data = AttrMapValue(isNoneOrListOfNoneOrNumbers, desc="Numbers used on r/hand side of legend (or None)"),
        leftPadding = AttrMapValue(isNumber, desc='Padding on left of drawing'),
        rightPadding = AttrMapValue(isNumber, desc='Padding on right of drawing'),
        topPadding = AttrMapValue(isNumber, desc='Padding at top of drawing'),
        bottomPadding = AttrMapValue(isNumber, desc='Padding at bottom of drawing'),
        drawLegend2 = AttrMapValue(isBoolean, desc='if legend 2(percentage)'),
        drawLegend3 = AttrMapValue(isBoolean, desc='if legend 2(deviations)'),
        legend3Format = AttrMapValue(None, desc="Formatting routine for deviation on right hand side of legend."),
        legend3Offset = AttrMapValue(isNumber, desc="Horizontal space between percentage and deviation on r/hand side"),
        legend3_data = AttrMapValue(isNoneOrListOfNoneOrNumbers, desc="Deviations used on r/hand side of legend (or None)"),
        )
    
    def __init__(self):
        piecharts.Pie.__init__(self)
        self.x = 0
        self.y = 0
        self.height = 3.5*cm
        self.width = 3.5*cm
        #self.height = 4*cm
        #self.width = 7*cm
        self.data = [38.4, 20.7, 18.9, 15.4, 6.6]
        self.labels = None
        self.direction = 'clockwise'
        
        self.pieAndLegend_colors = [ChbkYellow, ChbkBlue,
                               ChbkYellow75, ChbkBlue75,
                               ChbkYellow50, ChbkBlue50]
        
        #Allows us up to six 'wedges' to be coloured
        self.slices[0].fillColor=self.pieAndLegend_colors[0]
        self.slices[1].fillColor=self.pieAndLegend_colors[1]
        self.slices[2].fillColor=self.pieAndLegend_colors[2]
        self.slices[3].fillColor=self.pieAndLegend_colors[3]
        self.slices[4].fillColor=self.pieAndLegend_colors[4]
        self.slices[5].fillColor=self.pieAndLegend_colors[5]
        
        self.slices.strokeWidth = 0.75
        self.slices.strokeColor = _PCMYK_black
        
        legendOffset = 0.6*cm
        self.legendNumberOffset = 1.8*cm
        self.legendNumberFormat = '%.1f%%'
        self.legend_data = self.data
        
        #set up the legends
        self.legend1 = Legend()
        self.legend1.x = self.width+legendOffset
        self.legend1.y = self.height
        self.legend1.deltax = 0.2*cm
        self.legend1.deltay = 0.5*cm
        self.legend1.dxTextSpace = 0.4*cm
        self.legend1.dx = 0.2*cm
        self.legend1.dy = 0.2*cm
        self.legend1.columnMaximum = 7
        self.legend1.alignment = 'right'
        self.legend_names = ['AAA:','AA:','A:','BBB:','NR:']
        for f in range(0,len(self.data)):
            self.legend1.colorNamePairs.append((self.pieAndLegend_colors[f], self.legend_names[f]))
        self.legend1.fontName = "Avenir-Light"
        self.legend1.fontSize = 6.5
        self.legend1.strokeColor = _PCMYK_black
        self.legend1.strokeWidth = 0.5
        
        self.drawLegend2 = 1
        self._legend2 = Legend()
        self._legend2.dxTextSpace = 0
        self._legend2.dx = 0
        self._legend2.alignment = 'right'
        self._legend2.fontName = "Avenir-Light"
        self._legend2.fontSize = 6.5
        self._legend2.strokeColor = self.legend1.strokeColor
        
        self.drawLegend3 = 0
        self.legend3_data =[2.4, -2.7, 1.9, 5.4, 0.6]
        self._legend3 = Legend()
        self._legend3.dxTextSpace = 0
        self._legend3.dx = 0
        self._legend3.alignment = 'right'
        self._legend3.fontName = "Avenir-Light"
        self._legend3.fontSize = 6.5
        self._legend3.strokeColor = self.legend1.strokeColor
        self.legend3Offset = 0.8*cm
        self.legend3Format = '(%.1f%%)'
        
        self.leftPadding = 5
        self.rightPadding = 5
        self.topPadding = 5
        self.bottomPadding = 5
        self.drawLegend = 1

    
    def draw(self):
        if self.drawLegend:
            self.legend1.colorNamePairs = []
            self._legend2.colorNamePairs = []
            self._legend3.colorNamePairs = []
        for f in range(0,len(self.data)):
            if self.legend_names == None:
                self.slices[f].fillColor = self.pieAndLegend_colors[f]
                self.legend1.colorNamePairs.append((self.pieAndLegend_colors[f], None))
            else:
                try:
                    self.slices[f].fillColor = self.pieAndLegend_colors[f]
                    self.legend1.colorNamePairs.append((self.pieAndLegend_colors[f], self.legend_names[f]))
                except IndexError:
                    self.slices[f].fillColor = self.pieAndLegend_colors[f%len(self.pieAndLegend_colors)]
                    self.legend1.colorNamePairs.append((self.pieAndLegend_colors[f%len(self.pieAndLegend_colors)], self.legend_names[f]))
            if self.drawLegend2:
                _addLegendData(self._legend2,f,self.legend_data,self.legendNumberFormat)
            if self.drawLegend3:
                _addLegendData(self._legend3,f,self.legend3_data,self.legend3Format)
        
        p = piecharts.Pie.draw(self)
        if self.drawLegend:
            p.add(self.legend1)
            x = self.legend1.x
            if self.drawLegend2:
                #hide from user - keeps both sides lined up!
                x = self._legend2.x = x+self.legendNumberOffset
                self._legend2.y = self.legend1.y
                self._legend2.deltax = self.legend1.deltax
                self._legend2.deltay = self.legend1.deltay
                self._legend2.dy = self.legend1.dy
                self._legend2.columnMaximum = self.legend1.columnMaximum
                p.add(self._legend2)
            if self.drawLegend3:
                #hide from user - keeps both sides lined up!
                self._legend3.x = x+self.legend3Offset
                self._legend3.y = self.legend1.y
                self._legend3.deltax = self.legend1.deltax
                self._legend3.deltay = self.legend1.deltay
                self._legend3.dy = self.legend1.dy
                self._legend3.columnMaximum = self.legend1.columnMaximum
                p.add(self._legend3)
        p.shift(self.leftPadding, self.bottomPadding)
        return p
    
    def _getDrawingDimensions(self):
        tx = self.rightPadding
        if self.drawLegend:
            tx = tx+self.legend1.x+self.legendNumberOffset #self._legend2.x
            tx = tx + self._legend2._calculateMaxWidth(self._legend2.colorNamePairs)
        ty = self.bottomPadding+self.height+self.topPadding
        return (tx,ty)
    
    def demo(self, drawing=None):
        if not drawing:
            tx,ty = self._getDrawingDimensions()
            drawing = Drawing(tx, ty)
        drawing.add(self.draw())
        return drawing


class DerivativePieChartBase(DRRDrawing):
    def __init__(self,width=400,height=200,*args,**kw):
        DRRDrawing.__init__(self,width,height,*args,**kw)
        self._add(self,PieChart(),name='PieChart',validate=None,desc='The main chart')
    def applyDataSet(self, data):
        """Pull out and map the data manually
        
        This replaces all the 'associations' stuff we would normally use. The
        standard 'rowmap' association is not good at filtering empty or negative
        rows.  For this chart we always have 10 columns of names and values in
        the data file but several will be null; AND negative values are to be
        discarded.
        
        Doing this with one-liners in the chart body in getContents is ultra-painful,
        as the drawing editor does not let you do indented code.
        
        """
        self.chartId = data[0][0]
        numbers = []
        labels = []
        for i in range(10):
            label = data[0][2*i + 1]
            num = data[0][2*i + 2]
            if num is None:
                pass
            elif num < 0:
                print('negative slice discarded: %s, %0.2f' % (label, num))
            else:
                numbers.append(num)
                labels.append(label)
        self.PieChart.data = numbers
        self.PieChart.legend_names = labels
        self.PieChart.legend_data = numbers

## DRAWINGS have been moved into their own files

if __name__=='__main__':
    from reportlab.graphics import renderPDF, renderPM
    from sampledata import renderPS_SEP
    def render(d, name):
        d.save(fnRoot=name, formats=['pdf'], outDir='./output', verbose=1)
    
    #drawings all moved to their own files - only demos here are now widgets.
    
    # loops over all classes, uses render function(above)
    # (Now also saves in output directory)
    for C in [DotBox, SlideBox, ScatterPlot, HorizontalBarChart,
     HorizontalBarChartWCentreAxis, VerticalBarChart, LineChart,
     PieChart]:
        render(C().demo(), C.__name__)
